import logging

import requests
from requests.adapters import HTTPAdapter

from zenpy.lib.api import UserApi, Api, TicketApi, OrganizationApi, SuspendedTicketApi, EndUserApi, TicketImportAPI, \
    RequestAPI, OrganizationMembershipApi, AttachmentApi, SharingAgreementAPI, SatisfactionRatingApi, MacroApi
from zenpy.lib.cache import ZenpyCache
from zenpy.lib.endpoint import Endpoint
from zenpy.lib.exception import ZenpyException
from zenpy.lib.manager import ClassManager, ObjectManager

log = logging.getLogger()

__author__ = 'facetoe'


class Zenpy(object):
    """"""

    DEFAULT_TIMEOUT = 60.0

    @staticmethod
    def http_adapter_kwargs():
        """
        Provides Zenpy's default HTTPAdapter args for those users providing their own adapter.
        """

        return dict(
            # http://docs.python-requests.org/en/latest/api/?highlight=max_retries#requests.adapters.HTTPAdapter
            max_retries=3
        )

    def __init__(self, subdomain, email=None, token=None, oauth_token=None, password=None, session=None, timeout=None):
        """
        Python Wrapper for the Zendesk API.

        There are several ways to authenticate with the Zendesk API:
            * Email and password
            * Email and Zendesk API token
            * Email and OAuth token
            * Existing authenticated Requests Session object.


        :param subdomain: your Zendesk subdomain
        :param email: email address
        :param token: Zendesk API token
        :param oauth_token: OAuth token
        :param password: Zendesk password
        :param session: Existing Requests Session object
        :param timeout: Global timeout on API requests.
        """

        session = self._init_session(email, token, oauth_token, password, session)

        timeout = timeout or self.DEFAULT_TIMEOUT
        endpoint = Endpoint()

        self.users = UserApi(
            subdomain=subdomain,
            session=session,
            timeout=timeout,
            endpoint=endpoint.users)

        self.user_fields = Api(
            subdomain=subdomain,
            session=session,
            timeout=timeout,
            endpoint=endpoint.user_fields,
            object_type='user_field'
        )

        self.groups = Api(
            subdomain=subdomain,
            session=session,
            timeout=timeout,
            endpoint=endpoint.groups,
            object_type='group')

        self.macros = MacroApi(
            subdomain=subdomain,
            session=session,
            timeout=timeout,
        )

        self.organizations = OrganizationApi(
            subdomain=subdomain,
            session=session,
            timeout=timeout,
            endpoint=endpoint.organizations)

        self.organization_memberships = OrganizationMembershipApi(
            subdomain=subdomain,
            session=session,
            timeout=timeout,
            endpoint=endpoint.organization_memberships
        )

        self.tickets = TicketApi(
            subdomain=subdomain,
            session=session,
            timeout=timeout,
            endpoint=endpoint.tickets)

        self.suspended_tickets = SuspendedTicketApi(
            subdomain=subdomain,
            session=session,
            timeout=timeout,
            object_type='suspended_ticket',
            endpoint=endpoint.suspended_tickets)

        self.search = Api(
            subdomain=subdomain,
            session=session,
            timeout=timeout,
            endpoint=endpoint.search,
            object_type='results')

        self.topics = Api(
            subdomain=subdomain,
            session=session,
            timeout=timeout,
            endpoint=endpoint.topics,
            object_type='topic')

        self.attachments = AttachmentApi(
            subdomain=subdomain,
            session=session,
            timeout=timeout,
            endpoint=endpoint.attachments)

        self.brands = Api(
            subdomain=subdomain,
            session=session,
            timeout=timeout,
            endpoint=endpoint.brands,
            object_type='brand')

        self.job_status = Api(
            subdomain=subdomain,
            session=session,
            timeout=timeout,
            endpoint=endpoint.job_statuses,
            object_type='job_status')

        self.tags = Api(
            subdomain=subdomain,
            session=session,
            timeout=timeout,
            endpoint=endpoint.tags,
            object_type='tag')

        self.satisfaction_ratings = SatisfactionRatingApi(
            subdomain=subdomain,
            session=session,
            timeout=timeout,
            endpoint=endpoint.satisfaction_ratings
        )

        self.sharing_agreements = SharingAgreementAPI(
            subdomain=subdomain,
            session=session,
            timeout=timeout,
            endpoint=endpoint.sharing_agreements,
        )

        self.activities = Api(
            subdomain=subdomain,
            session=session,
            timeout=timeout,
            endpoint=endpoint.activities,
            object_type='activity'
        )

        self.group_memberships = Api(
            subdomain=subdomain,
            session=session,
            timeout=timeout,
            endpoint=endpoint.group_memberships,
            object_type='group_membership'
        )

        self.end_user = EndUserApi(
            subdomain=subdomain,
            session=session,
            timeout=timeout,
            endpoint=endpoint.end_user
        )

        self.ticket_metrics = Api(
            subdomain=subdomain,
            session=session,
            timeout=timeout,
            endpoint=endpoint.ticket_metrics,
            object_type='ticket_metric'
        )

        self.ticket_fields = Api(
            subdomain=subdomain,
            session=session,
            timeout=timeout,
            endpoint=endpoint.ticket_fields,
            object_type='ticket_field'
        )

        self.ticket_import = TicketImportAPI(
            subdomain=subdomain,
            session=session,
            timeout=timeout,
            endpoint=endpoint.ticket_import
        )

        self.requests = RequestAPI(
            subdomain=subdomain,
            session=session,
            timeout=timeout,
            endpoint=endpoint.requests
        )

    def _init_session(self, email, token, oath_token, password, session):
        if not session:
            session = requests.Session()
            # Workaround for possible race condition - https://github.com/kennethreitz/requests/issues/3661
            session.mount('https://', HTTPAdapter(**self.http_adapter_kwargs()))

        if not hasattr(session, 'authorized') or not session.authorized:
            # session is not an OAuth session that has been authorized, so authorize the session.
            if not password and not token and not oath_token:
                raise ZenpyException("password, token or oauth_token are required!")
            elif password and token:
                raise ZenpyException("password and token "
                                     "are mutually exclusive!")
            if password:
                session.auth = (email, password)
            elif token:
                session.auth = ('%s/token' % email, token)
            elif oath_token:
                session.headers.update({'Authorization': 'Bearer %s' % oath_token})
            else:
                raise ZenpyException("Invalid arguments to _init_session()!")

        headers = {'Content-type': 'application/json',
                   'User-Agent': 'Zenpy/1.1'}
        session.headers.update(headers)
        return session

    def get_cache_names(self):
        """
        Returns a list of current caches
        """
        return self._get_cache_mapping().keys()

    def get_cache_max(self, cache_name):
        """
        Returns the maxsize attribute of the named cache
        """
        return self._get_cache(cache_name).maxsize

    def set_cache_max(self, cache_name, maxsize, **kwargs):
        """
        Sets the maxsize attribute of the named cache
        """
        cache = self._get_cache(cache_name)
        cache.set_maxsize(maxsize, **kwargs)

    def get_cache_impl_name(self, cache_name):
        """
        Returns the name of the cache implementation for the named cache
        """
        return self._get_cache(cache_name).impl_name

    def set_cache_implementation(self, cache_name, impl_name, maxsize, **kwargs):
        """
        Changes the cache implementation for the named cache
        """
        self._get_cache(cache_name).set_cache_impl(impl_name, maxsize, **kwargs)

    def add_cache(self, object_type, cache_impl_name, maxsize, **kwargs):
        """
        Add a new cache for the named object type and cache implementation
        """
        if object_type not in self._get_class_mapping():
            raise ZenpyException("No such object type: %s" % object_type)
        cache_mapping = self._get_cache_mapping()
        cache_mapping[object_type] = ZenpyCache(cache_impl_name, maxsize, **kwargs)

    def delete_cache(self, cache_name):
        """
        Deletes the named cache
        """
        cache_mapping = self._get_cache_mapping()
        del cache_mapping[cache_name]

    def _get_cache(self, cache_name):
        cache_mapping = self._get_cache_mapping()
        if cache_name not in cache_mapping:
            raise ZenpyException("No such cache - %s" % cache_name)
        else:
            return cache_mapping[cache_name]

    def _get_cache_mapping(self):
        return ObjectManager.cache_mapping

    def _get_class_mapping(self):
        return ClassManager.class_mapping



from setuptools import setup
import setuptools

setup(
    name='zenpy',
    packages=setuptools.find_packages(),
    version='1.1.9-1',
    description='Python wrapper for the Zendesk API',
    license='GPLv3',
    author='Face Toe',
    author_email='facetoe@facetoe.com.au',
    url='https://github.com/facetoe/zenpy',
    download_url='https://github.com/facetoe/zenpy/releases/tag/1.1.9-1',
    install_requires=[
        'requests>=2.7.0',
        'python-dateutil>=2.4.0',
        'cachetools>=1.0.3'
    ],
    keywords=['zendesk', 'api', 'wrapper'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)'
    ],
)



import logging
from datetime import datetime, date
from json import JSONEncoder

from zenpy.lib.api_objects import Activity, Request, UserRelated, OrganizationMembership, Upload, SharingAgreement, \
    Macro, Action, MacroResult
from zenpy.lib.api_objects import Attachment
from zenpy.lib.api_objects import Audit
from zenpy.lib.api_objects import Brand
from zenpy.lib.api_objects import CcEvent
from zenpy.lib.api_objects import ChangeEvent
from zenpy.lib.api_objects import Comment
from zenpy.lib.api_objects import CommentPrivacyChangeEvent
from zenpy.lib.api_objects import CreateEvent
from zenpy.lib.api_objects import ErrorEvent
from zenpy.lib.api_objects import ExternalEvent
from zenpy.lib.api_objects import FacebookCommentEvent
from zenpy.lib.api_objects import FacebookEvent
from zenpy.lib.api_objects import Group
from zenpy.lib.api_objects import GroupMembership
from zenpy.lib.api_objects import JobStatus
from zenpy.lib.api_objects import LogmeinTranscriptEvent
from zenpy.lib.api_objects import Metadata
from zenpy.lib.api_objects import NotificationEvent
from zenpy.lib.api_objects import Organization
from zenpy.lib.api_objects import OrganizationActivityEvent
from zenpy.lib.api_objects import OrganizationField
from zenpy.lib.api_objects import PushEvent
from zenpy.lib.api_objects import SatisfactionRating
from zenpy.lib.api_objects import SatisfactionRatingEvent
from zenpy.lib.api_objects import Source
from zenpy.lib.api_objects import Status
from zenpy.lib.api_objects import SuspendedTicket
from zenpy.lib.api_objects import System
from zenpy.lib.api_objects import Tag
from zenpy.lib.api_objects import Thumbnail
from zenpy.lib.api_objects import Ticket
from zenpy.lib.api_objects import TicketAudit
from zenpy.lib.api_objects import TicketEvent
from zenpy.lib.api_objects import TicketField
from zenpy.lib.api_objects import TicketMetric
from zenpy.lib.api_objects import TicketMetricItem
from zenpy.lib.api_objects import TicketSharingEvent
from zenpy.lib.api_objects import Topic
from zenpy.lib.api_objects import TweetEvent
from zenpy.lib.api_objects import User
from zenpy.lib.api_objects import UserField
from zenpy.lib.api_objects import Via
from zenpy.lib.api_objects import VoiceCommentEvent
from zenpy.lib.cache import ZenpyCache
from zenpy.lib.exception import ZenpyException
from zenpy.lib.util import to_snake_case

log = logging.getLogger(__name__)

__author__ = 'facetoe'


class ApiObjectEncoder(JSONEncoder):
    """ Class for encoding API objects"""

    def default(self, o):
        if hasattr(o, 'to_dict'):
            return o.to_dict()
        elif isinstance(o, datetime):
            return o.date().isoformat()
        elif isinstance(o, date):
            return o.isoformat()


class ClassManager(object):
    """
    ClassManager provides methods for converting JSON objects
    to the correct Python ones.
    """

    class_mapping = {
        'ticket': Ticket,
        'user': User,
        'organization': Organization,
        'group': Group,
        'brand': Brand,
        'topic': Topic,
        'comment': Comment,
        'attachment': Attachment,
        'thumbnail': Thumbnail,
        'metadata': Metadata,
        'system': System,
        'create': CreateEvent,
        'change': ChangeEvent,
        'notification': NotificationEvent,
        'voicecomment': VoiceCommentEvent,
        'commentprivacychange': CommentPrivacyChangeEvent,
        'satisfactionrating': SatisfactionRatingEvent,
        'ticketsharingevent': TicketSharingEvent,
        'organizationactivity': OrganizationActivityEvent,
        'error': ErrorEvent,
        'tweet': TweetEvent,
        'facebookevent': FacebookEvent,
        'facebookcomment': FacebookCommentEvent,
        'external': ExternalEvent,
        'logmeintranscript': LogmeinTranscriptEvent,
        'push': PushEvent,
        'cc': CcEvent,
        'via': Via,
        'source': Source,
        'job_status': JobStatus,
        'audit': Audit,
        'ticket_event': TicketEvent,
        'tag': Tag,
        'suspended_ticket': SuspendedTicket,
        'ticket_audit': TicketAudit,
        'satisfaction_rating': SatisfactionRating,
        'activity': Activity,
        'group_membership': GroupMembership,
        'ticket_metric': TicketMetric,
        'ticket_metric_event': TicketMetric,
        'status': Status,
        'ticket_metric_item': TicketMetricItem,
        'user_field': UserField,
        'organization_field': OrganizationField,
        'ticket_field': TicketField,
        'request': Request,
        'user_related': UserRelated,
        'organization_membership': OrganizationMembership,
        'upload': Upload,
        'sharing_agreement': SharingAgreement,
        'macro': Macro,
        'action': Action,
        'result': MacroResult
    }

    def __init__(self, api):
        self.api = api

    def object_from_json(self, object_type, object_json):
        obj = self.class_for_type(object_type)
        return self._object_from_json(obj, object_json)

    def class_for_type(self, object_type):
        if object_type not in self.class_mapping:
            raise ZenpyException("Unknown object_type: " + str(object_type))
        else:
            return self.class_mapping[object_type]

    def _object_from_json(self, object_type, object_json):
        # This method is recursive, if we have already
        # created this object just return it.
        if not isinstance(object_json, dict):
            return object_json

        obj = object_type(api=self.api)
        for key, value in object_json.items():
            if key in ('metadata', 'from', 'system', 'photo', 'thumbnails'):
                key = '_%s' % key

            if key in self.class_mapping.keys():
                value = self.object_from_json(key, value)
            setattr(obj, key, value)
        return obj


class ObjectManager(object):
    """
    The ObjectManager is responsible for maintaining various caches
    and also provides access to the ClassManager
    """

    cache_mapping = {
        'user': ZenpyCache('LRUCache', maxsize=10000),
        'organization': ZenpyCache('LRUCache', maxsize=10000),
        'group': ZenpyCache('LRUCache', maxsize=10000),
        'brand': ZenpyCache('LRUCache', maxsize=10000),
        'ticket': ZenpyCache('TTLCache', maxsize=10000, ttl=30),
        'comment': ZenpyCache('LRUCache', maxsize=10000),
        'request': ZenpyCache('LRUCache', maxsize=10000),
        'user_field': ZenpyCache('TTLCache', maxsize=10000, ttl=30),
        'organization_field': ZenpyCache('LRUCache', maxsize=10000),
        'ticket_field': ZenpyCache('LRUCache', maxsize=10000),
        'sharing_agreement': ZenpyCache('TTLCache', maxsize=10000, ttl=6000),
    }

    def __init__(self, api):
        self.class_manager = ClassManager(api)

    def object_from_json(self, object_type, object_json):
        return self.class_manager.object_from_json(object_type, object_json)

    def delete_from_cache(self, obj):
        if isinstance(obj, list):
            for o in obj:
                self._delete_from_cache(o)
        else:
            self._delete_from_cache(obj)

    def _delete_from_cache(self, obj):
        object_type = to_snake_case(obj.__class__.__name__)
        if object_type in self.cache_mapping:
            cache = self.cache_mapping[object_type]
            obj = cache.pop(obj.id, None)
            if obj:
                log.debug("Cache RM: [%s %s]" % (object_type.capitalize(), obj.id))

    def query_cache(self, object_type, _id):
        if object_type not in self.cache_mapping.keys():
            return None

        cache = self.cache_mapping[object_type]
        if _id in cache:
            log.debug("Cache HIT: [%s %s]" % (object_type.capitalize(), _id))
            return cache[_id]
        else:
            log.debug('Cache MISS: [%s %s]' % (object_type.capitalize(), _id))

    def update_caches(self, _json):
        if 'results' in _json:
            self._cache_search_results(_json)
        else:
            for object_type in self.cache_mapping.keys():
                self._add_to_cache(object_type, _json)

    def _add_to_cache(self, object_type, object_json):
        cache = self.cache_mapping[object_type]
        multiple_key = object_type + 's'
        if object_type in object_json:
            obj = object_json[object_type]
            log.debug("Caching: [%s %s]" % (object_type.capitalize(), obj['id']))
            self._cache_item(cache, obj, object_type)

        elif multiple_key in object_json:
            objects = object_json[multiple_key]
            log.debug("Caching %s %s " % (len(objects), multiple_key.capitalize()))
            for obj in object_json[multiple_key]:
                self._cache_item(cache, obj, object_type)

    def _cache_search_results(self, _json):
        results = _json['results']
        log.debug("Caching %s search results" % len(results))
        for result in results:
            object_type = result['result_type']
            cache = self.cache_mapping[object_type]
            self._cache_item(cache, result, object_type)

    def _cache_item(self, cache, item_json, item_type):
        key = self.get_key(item_type)
        cache[item_json[key]] = self.object_from_json(item_type, item_json)

    def get_key(self, item_type):
        if item_type in ('user_field', 'organization_field'):
            key = 'key'
        else:
            key = 'id'
        return key



import re

FIRST_CAP_REGEX = re.compile('(.)([A-Z][a-z]+)')
ALL_CAP_REGEX = re.compile('([a-z0-9])([A-Z])')


def to_snake_case(name):
    s1 = FIRST_CAP_REGEX.sub(r'\1_\2', name)
    return ALL_CAP_REGEX.sub(r'\1_\2', s1).lower()


def is_timezone_aware(datetime_obj):
    """
    Determine whether or not a given datetime object is timezone aware.
    """
    return datetime_obj.tzinfo is not None and datetime_obj.tzinfo.utcoffset(datetime_obj) is not None


def is_iterable_but_not_string(obj):
    """
    Determine whether or not obj is iterable but not a string (eg, a list, set, tuple etc).
    """
    return hasattr(obj, '__iter__') and not isinstance(obj, str) and not isinstance(obj, bytes)
